#!/usr/bin/python

"""Miscellaneous things that should come into the genesilico.structure.
   save() - saves structure with HEADER, END and SEQRES records.
"""

from Bio.PDB       import to_one_letter_code, PDBParser
from Bio.PDB.PDBIO import PDBIO
import os.path

def save_structure(structure, filename, headers = []):
  """Writes a structure to PDB file, along with all metainformation.
     Uses Bio.PDBIO, adds HEADER, SEQRES, END records and any headers
     that are explicitly given.
     structure -- Bio.PDB.Structure
     filename   -- string with output filename or Python file object
     headers   -- list of one-line strings without EOLs.
     Returns: nothing.
  """
  if type(filename) == type(''):
    filename = open(filename, "w")
    to_be_closed = True
  else:
    to_be_closed = False
  model = structure[0]
  print >>filename, "HEADER      %-50s%4s" % (structure.id, structure.id[:4])
  for header in headers:
    print >>filename, header
  # write SEQRES
  for chain in model:
    residue_names  = []
    current_resid  = 0
    for residue in chain:
      if residue.id[0] != ' ': # water erroneously put into the main chain
        continue
      while current_resid + 1 < residue.get_id()[1]:
        residue_names.append("GLY") # Don't know the real sequence...
        current_resid += 1
      residue_names.append(residue.resname)
      current_resid += 1
    total_len = len(residue_names)
    i = 1
    residue_codes_per_line = 13
    while residue_names != []:   
      print >>filename, ("SEQRES %3d %s %4d  %s" %
                        (i, chain.id, total_len,
                         ' '.join(residue_names[:residue_codes_per_line])))
      residue_names = residue_names[residue_codes_per_line:]
      i += 1
  pdbio = PDBIO()
  pdbio.set_structure(structure)
  pdbio.save(filename)
  print >>filename, "END"
  if to_be_closed:
    filename.close()

def modelname_to_pdbid(filename):
  """Converts filename in the format used by FRankenstein into PDB Id of the template.
     filename -- string
     Returns: string.
  """
  modelname, _ = os.path.splitext(os.path.basename(filename))
  if modelname.startswith("pdb"):
    modelname = modelname[3:]
  if modelname.endswith('-Verify3D'):
    modelname = modelname[:-9]
  name = modelname[modelname.rfind("-")+1:]
  if name[5:6].isdigit() or name[5:6] == '_':
    name = name[:4]
  if name[-2:] == ".c":
    name = name[:-2]
#  if name[-2:] == "_0": # WORKAROUND: misnamed Blast output from metaserver
#    name = name[:-2]
  return name[:6]

def load_structure(filename):
  """Loads PDB structure.
     filename -- string,
     Returns: Bio.PDB.Structure.
  """
  import genesilico.structure.BrutePDBParser 
  pdbid = modelname_to_pdbid(filename)
  parser = genesilico.structure.BrutePDBParser.BrutePDBParser(PERMISSIVE=2)
  return parser.get_structure(pdbid, filename)

class All(object):
  def __contains__(self, obj):
    return True

def chain2sequence(chain, allowed_atom_types            = All(),
                          keep_numbering                = True,
                          check_if_residues_consecutive = True):
  sequence = []
  aa_num = 0
  for residue in chain.get_iterator():
    for atom in iter(residue):
      if atom.get_name() in allowed_atom_types:
        try:
          aa = to_one_letter_code[residue.resname]
        except KeyError:
          aa = 'X'
        if keep_numbering:
          if check_if_residues_consecutive:
            assert aa_num < residue.get_id()[1]
          while aa_num < residue.get_id()[1] - 1:
            sequence.append(".")
            aa_num += 1
          if check_if_residues_consecutive:
            assert aa_num == residue.get_id()[1] - 1
          aa_num = residue.get_id()[1]
        sequence.append(aa)
        break
  sequence = ''.join(sequence)
  return sequence

def structure2sequences(structure, allowed_atom_types = All(),
                                   keep_numbering     = True,
                                   check_if_residues_consecutive = True):
  if type(structure)==str:
    parser = PDBParser(PERMISSIVE=1)
    try:
      structure = parser.get_structure(structure, structure)
    except IOError:
      return None
  result = []
  for model in structure.get_iterator():
    for chain in model.get_iterator():
      sequence = chain2sequence(chain, allowed_atom_types            = allowed_atom_types,
                                       keep_numbering                = keep_numbering,
                                       check_if_residues_consecutive = check_if_residues_consecutive )
      result.append((chain.id, sequence))
  return result

def structure2sequence(structure, allowed_atom_types            = All(),
                                  chain_id                      = None,
                                  keep_numbering                = True,
                                  check_if_residues_consecutive = True):
  result = structure2sequences(structure, allowed_atom_types = allowed_atom_types,
                                          keep_numbering = keep_numbering,
                                          check_if_residues_consecutive = check_if_residues_consecutive)
  if chain_id != None:
    return dict(result)[chain_id]
  else:
    return result[0][1]


